/* DemoXML.cpp : dmonstration des techniques pour srialiser en XML avec UtilXML.
 *     
 * Auteur : M.M.
 * Cration : fvrier 2007, modif. : fvrier 2009
 */
#include <iostream>
#include <iomanip>
#include <vector>
#include "UtilXML.h"
using namespace std;
using namespace UtilXML;

// On veut une identification quelconque permettant de dtecter les versions de fichier 
// de donnes incompatibles. On devrait changer la valeur  chaque modification des
// structures des donnes srialises.
const string VERSION_DMO= "1.0";

// Voici le nom du fichier qui contiendra les donnes en XML. En principe, il devrait
// porter le suffixe "xml".
const string FICHIER_DMO= "Dmo.xml";

/////////////////////////////////////////////////////////////////////////////////////////
// SRIALISATION DES ENUM
// - aucune fonction de srialisation personnelle  faire;
// - si on a une fonction de conversion en texte, on peut l'utiliser pour faire crire la
//   valeur en texte dans le fichier XML si a peut aider au dbogage.
//
enum TypeMois { JANVIER=1, FVRIER, MARS, AVRIL, MAI, JUIN,
			    JUILLET, AOT, SEPTEMBRE, OCTOBRE, NOVEMBRE, DCEMBRE };

enum TypeVaritPomme { MCINTOSH, LOBO, CORTLAND };

std::string VaritPommeEnTexte(TypeVaritPomme p_varit)
    {
    string nomVarit;  // Donnera le nom officiel de la varit de pommes

    switch (p_varit)
        {
        case MCINTOSH : nomVarit= "McIntosh"; break;
        case LOBO     : nomVarit= "Lobo";     break;
        case CORTLAND : nomVarit= "Cortland"; break;
        default : assert(false);
        }

    return nomVarit;
    }

/////////////////////////////////////////////////////////////////////////////////////////
// SRIALISATION DES STRUCT
// - on doit faire des fonctions pour la srialisation et la dsrialisation;
// - les fonctions doivent avoir trois paramtres, dans cet ordre : le fichier, le nom du
//   champ et la valeur  srialiser ou dsrialiser;
// - le paramtre pour la valeur doit videmment tre un paramtre de sortie dans la
//   dsrialisation;
// - les fonctions doivent utiliser les fonctions de UtilXML ou les fonctions personnelles
//   de srialisation/dsrialisation pour les structures imbriques;
// - pour la srialisation, il faut commencer par appeler crireBaliseDbutStruct, puis
//   srialiser les champs, puis appeler crireBaliseFin;
// - pour la dsrialisation, il faut commencer par appeler VrifierBaliseDbutStruct,
//   puis dsrialiser les champs, puis appeler VrifierBaliseFin.
//
// Exemple 1 : Srialisation du TypeDate
//
// - pour les types de base, il suffit d'appeler la bonne fonction et passer le fichier,
//   le nom du champ (qui deviendra la balise) et la valeur;
// - pour le type enum, le troisime paramtre est plutt le nom du type, la valeur est en
//   quatrime. Pour la srialisation du mois, on n'a pas utilis le cinquime paramtre
//   qui aurait permis de passer une valeur texte, donc seule la valeur numrique sera
//   inscrite dans le fichier (on considre que le numro de mois est assez significatif).
//
struct TypeDate
	{
	int jour;
	TypeMois mois;
	int anne;
	};

void SrialiserDate(ofstream& p_es_fic, const string& p_nomChamp, const TypeDate& p_date)
    {
	crireBaliseDbutStruct(p_es_fic, p_nomChamp, "TypeDate");
	SrialiserInt(p_es_fic, "jour", p_date.jour);
	SrialiserEnum(p_es_fic, "mois", "TypeMois", p_date.mois);
	SrialiserInt(p_es_fic, "anne", p_date.anne);
	crireBaliseFin(p_es_fic, p_nomChamp);
    }

void DsrialiserDate(ifstream& p_es_fic, const string& p_nomChamp, TypeDate& p_s_date)
    {
	VrifierBaliseDbutStruct(p_es_fic, p_nomChamp, "TypeDate");
	DsrialiserInt(p_es_fic, "jour", p_s_date.jour);
	DsrialiserEnum(p_es_fic, "mois", "TypeMois", p_s_date.mois);
    DsrialiserInt(p_es_fic, "anne", p_s_date.anne);
	VrifierBaliseFin(p_es_fic, p_nomChamp);
    }

// Exemple 2 : Srialisation du TypeConsommationJourne
//
// - pour la date, on appelle directement les fonctions qu'on vient de faire, comme les
//   autres fonctions de UtilXML;
// - pour la srialisation de la varit, un type enum, on a pass le dernier paramtre
//   permettant de mettre la valeur texte dans le fichier en plus de la valeur numrique;
// - pour les vecteurs, il faut passer le nom du type des lments en troisime paramtre
//   et, en dernier paramtre, la fonction de srialisation ou de dsrialisation pour
//   ces lments (ici on a pass SrialiserString de UtilXML, car les lments sont de
//   simples std::string. Voir Enregistrer/Rcuprer plus loin pour d'autres exemples). 
//
struct TypeConsommationJourne
    {
    TypeDate date;
	bool jene;
    int nbPommes;
    TypeVaritPomme varitPomme;
	std::vector<std::string> repasPrincipaux;
	std::string nomLgumePrincipal;
	double nbKgLgumePrincipal;
	};

void SrialiserConsommationJourne(ofstream& p_es_fic, const string& p_nomChamp,
			                       const TypeConsommationJourne& p_consommation)
    {
	crireBaliseDbutStruct(p_es_fic, p_nomChamp, "TypeConsommationJourne");
	SrialiserDate(p_es_fic, "date", p_consommation.date);
	SrialiserBool(p_es_fic, "jene", p_consommation.jene);
	SrialiserInt(p_es_fic, "nbPommes", p_consommation.nbPommes);
    SrialiserEnum(p_es_fic, "varitPomme", "TypeVaritPomme",
		           p_consommation.varitPomme,
		           VaritPommeEnTexte(p_consommation.varitPomme));
	SrialiserVector(p_es_fic, "repasPrincipaux", "std::string", 
		             p_consommation.repasPrincipaux, SrialiserString);
	SrialiserString(p_es_fic, "nomLgumePrincipal", p_consommation.nomLgumePrincipal);
	SrialiserDouble(p_es_fic, "nbKgLgumePrincipal", p_consommation.nbKgLgumePrincipal);
	crireBaliseFin(p_es_fic, p_nomChamp);
    }

void DsrialiserConsommationJourne(std::ifstream& p_es_fic, const string& p_nomChamp,
									 TypeConsommationJourne& p_s_consommation)
    {
	VrifierBaliseDbutStruct(p_es_fic, p_nomChamp, "TypeConsommationJourne");
	DsrialiserDate(p_es_fic, "date", p_s_consommation.date);
	DsrialiserBool(p_es_fic, "jene", p_s_consommation.jene);
	DsrialiserInt(p_es_fic, "nbPommes", p_s_consommation.nbPommes);
    DsrialiserEnum(p_es_fic, "varitPomme", "TypeVaritPomme",
		             p_s_consommation.varitPomme);
	DsrialiserVector(p_es_fic, "repasPrincipaux", "std::string",
		               p_s_consommation.repasPrincipaux, DsrialiserString);
	DsrialiserString(p_es_fic, "nomLgumePrincipal", p_s_consommation.nomLgumePrincipal);
	DsrialiserDouble(p_es_fic, "nbKgLgumePrincipal", p_s_consommation.nbKgLgumePrincipal);
	VrifierBaliseFin(p_es_fic, p_nomChamp);
    }

/////////////////////////////////////////////////////////////////////////////////////////
// FONCTIONS GNRALES DE SRIALISATION/DSRIALISATION
// - on commence la srialisation par crireEntteXML, si on veut un fichier standardis
//   et qu'on pourra ouvrir aisment dans Internet Explorer ou Visual Studio;
// - on poursuit en mettant une balise globale (racine) avec crireBaliseDbutFichier qui
//   permet aussi de mettre un numro de version;
// - on utilise ensuite les mmes techniques que pour la srialisation des champs
//   des struct;
// - pour la dsrialisation, on doit utiliser VrifierEntteXML au besoin, puis
//   VrifierBaliseDbutFichier. On poursuit avec les fonctions normales.
// - dans cet exemple, on doit srialiser/dsrialiser un vector<TypeConsommationJourne>,
//   donc on passe une fonction permettant de srialiser/dsrialiser ses lments,
//   c'est--dire une des fonctions que nous avons faites ci-dessus;
// - on doit aussi srialiser un vector<int> et on passe des fonctions de UtilXML;
// - la dsrialisation contient un try-catch parce que les erreurs dtectes par les
//   fonctions de UtilXML font lever une exception de type UtilXML::TypeErreurXML.
//
void EnregistrerInfoDmo(const string& p_nomFichier, 
						 const vector<TypeConsommationJourne>& p_consommations,
						 const vector<int>& p_diversEntiers,
						 const string& p_identification)
    {
    ofstream ficDmo(p_nomFichier.c_str());

	crireEntteXML(ficDmo);
	crireBaliseDbutFichier(ficDmo, "dmo", VERSION_DMO);
	SrialiserVector(ficDmo, "consommations", "TypeConsommationJourne", 
		             p_consommations, SrialiserConsommationJourne);
	SrialiserVector(ficDmo, "diversEntiers", "int", p_diversEntiers, SrialiserInt);
	SrialiserString(ficDmo, "identification", p_identification);
	crireBaliseFin(ficDmo, "dmo");
    }

bool RcuprerInfoDmo(const string& p_nomFichier, 
					   vector<TypeConsommationJourne>& p_s_consommations,
					   vector<int>& p_s_diversEntiers,
					   string& p_s_identification)
    {
    ifstream ficDmo(p_nomFichier.c_str());

	if (! ficDmo.is_open())
		return false;

	try
		{
		VrifierEntteXML(ficDmo);
		VrifierBaliseDbutFichier(ficDmo, "dmo", VERSION_DMO);
		DsrialiserVector(ficDmo, "consommations", "TypeConsommationJourne",
			               p_s_consommations, DsrialiserConsommationJourne);
		DsrialiserVector(ficDmo, "diversEntiers", "int", p_s_diversEntiers,
			               DsrialiserInt);
		DsrialiserString(ficDmo, "identification", p_s_identification);
		VrifierBaliseFin(ficDmo, "dmo");
		}
	catch (TypeErreurXML)
		{
		return false;
		}

	return true;
	}

// Programme principal qui simule des oprations sans tre interactif... Chaque excution
// ajoutera des donnes (jusqu' la troisime fois)
int main()
    {
	vector<TypeConsommationJourne> consommations;  // Les consommations quotidiennes
	vector<int> diversEntiers; // C'est bien a, juste un exemple...
	string identification;     // Idem...

	bool russi= RcuprerInfoDmo(FICHIER_DMO, consommations, diversEntiers, identification);

    // Affichage
	if (! russi)
		cout << "Aucune donne n'a t rcupre.\n";
	else
		{
		cout << "Identification : " << identification << '\n';

		for (int i= 0; i != consommations.size(); ++i)
			{
			cout << "consommations[" << i << "] :\n"
				 << "\tdate : " << consommations[i].date.anne << '-' 
				 << setfill('0') << right 
				 << setw(2) << static_cast<int>(consommations[i].date.mois) << '-' 
				 << setw(2) << consommations[i].date.jour
				 << setfill(' ') << '\n';

			if (consommations[i].jene)
				cout << "\tEn jeune\n";
			else
				{
				cout << "\tnb de pommes : " << consommations[i].nbPommes 
					 << " (" << VaritPommeEnTexte(consommations[i].varitPomme) << ")\n";

				const vector<string>& repasPrincipaux= consommations[i].repasPrincipaux;

				for (int j= 0; j != repasPrincipaux.size(); ++j)
					cout << "\tRepas " << j << " : " << repasPrincipaux[j] << '\n';

				cout << "\tLgume principal : " << consommations[i].nomLgumePrincipal 
					 << " (" << fixed << setprecision(2) 
					 << consommations[i].nbKgLgumePrincipal << " kg)\n";
				}
			}

		for (int i= 0; i != diversEntiers.size(); ++i)
			cout << "diversEntiers[" << i << "] : " << diversEntiers[i] << '\n';
		}

    // Ajout d'information (en trois tapes)
	if (!russi)
        identification= "MM";
	else
	if (diversEntiers.empty())
		{
		diversEntiers.push_back(123);
		diversEntiers.push_back(456);

		DsactiverFormatage(); // Pour dmo seulement, on ne ferait pas a normalement
		}
	else
	if (consommations.empty())
		{
		TypeConsommationJourne cons;
		cons.date.anne= 2007;
		cons.date.mois= FVRIER;
		cons.date.jour= 24;
		cons.jene= false;
		cons.nomLgumePrincipal= "Carotte";
		cons.nbKgLgumePrincipal= 0.2;
		cons.nbPommes= 2;
		cons.varitPomme= CORTLAND;
		cons.repasPrincipaux.push_back("Toast et caf");
		cons.repasPrincipaux.push_back("Macaroni <\" l'italienne\">"); // Pour tester encodage
		cons.repasPrincipaux.push_back("Fish & Chip");                  // ...
		consommations.push_back(cons);

		cons.date.jour= 25;
		cons.jene= true;
		cons.nbPommes= 0;
		cons.repasPrincipaux.clear();
		cons.nomLgumePrincipal.clear();
		cons.nbKgLgumePrincipal= 0;
		consommations.push_back(cons);

		ChangerIndentation(5); // Pour dmo seulement, on ne ferait pas a normalement
		}

	EnregistrerInfoDmo(FICHIER_DMO, consommations, diversEntiers, identification);
    }

/* 
-------------------------------------------------------------------------------
Rsultat ultime affich
-------------------------------------------------------------------------------
Identification : MM
consommations[0] :
        date : 2007-02-24
        nb de pommes : 2 (Cortland)
        Repas 0 : Toast et caf
        Repas 1 : Macaroni <" l'italienne">
        Repas 2 : Fish & Chip
        Lgume principal : Carotte (0.20 kg)
consommations[1] :
        date : 2007-02-25
        En jeune
diversEntiers[0] : 123
diversEntiers[1] : 456


-------------------------------------------------------------------------------
Fichier Dmo.xml aprs premire excution :
-------------------------------------------------------------------------------
<?xml version="1.0" encoding="ISO-8859-1" standalone="yes"?>
<dmo version="1.0">
   <consommations type="std::vector">
      <taille type="long">0</taille>
      <lments type="struct TypeConsommationJourne"></lments>
   </consommations>
   <diversEntiers type="std::vector">
      <taille type="long">0</taille>
      <lments type="int"></lments>
   </diversEntiers>
   <identification type="std::string">MM</identification>
</dmo>

-------------------------------------------------------------------------------
Fichier Dmo.xml ultime
-------------------------------------------------------------------------------
<?xml version="1.0" encoding="ISO-8859-1" standalone="yes"?>
<dmo version="1.0">
   <consommations type="std::vector">
      <taille type="long">2</taille>
      <lments type="struct TypeConsommationJourne">
         <lment type="struct TypeConsommationJourne">
            <date type="struct TypeDate">
               <jour type="int">24</jour>
               <mois type="enum TypeMois">2</mois>
               <anne type="int">2007</anne>
            </date>
            <jene type="bool">0</jene>
            <nbPommes type="int">2</nbPommes>
            <varitPomme type="enum TypeVaritPomme">2(Cortland)</varitPomme>
            <repasPrincipaux type="std::vector">
               <taille type="long">3</taille>
               <lments type="std::string">
                  <lment type="std::string">Toast et caf</lment>
                  <lment type="std::string">Macaroni &lt;&quot; l&apos;italienne&quot;&gt;</lment>
                  <lment type="std::string">Fish &amp; Chip</lment>
               </lments>
            </repasPrincipaux>
            <nomLgumePrincipal type="std::string">Carotte</nomLgumePrincipal>
            <nbKgLgumePrincipal type="double">2.0000000000000001e-001</nbKgLgumePrincipal>
         </lment>
         <lment type="struct TypeConsommationJourne">
            <date type="struct TypeDate">
               <jour type="int">25</jour>
               <mois type="enum TypeMois">2</mois>
               <anne type="int">2007</anne>
            </date>
            <jene type="bool">1</jene>
            <nbPommes type="int">0</nbPommes>
            <varitPomme type="enum TypeVaritPomme">2(Cortland)</varitPomme>
            <repasPrincipaux type="std::vector">
               <taille type="long">0</taille>
               <lments type="std::string"></lments>
            </repasPrincipaux>
            <nomLgumePrincipal type="std::string"></nomLgumePrincipal>
            <nbKgLgumePrincipal type="double">0.0000000000000000e+000</nbKgLgumePrincipal>
         </lment>
      </lments>
   </consommations>
   <diversEntiers type="std::vector">
      <taille type="long">2</taille>
      <lments type="int">
         <lment type="int">123</lment>
         <lment type="int">456</lment>
      </lments>
   </diversEntiers>
   <identification type="std::string">MM</identification>
</dmo>
*/
