/* UtilXML.cpp : fonctions d'assistance  la srialisation en format XML. Voir les
 *     commentaires dans le fichier d'en-tte ou la documentation pour plus de dtails
 *     sur l'utilisation ou la structure des fichiers crs.
 *
 *     N.B. En autant que possible, les principes, le code et sa structuration ont t
 *          limits pour tre comprhensibles avant la connaissance des classes et des
 *          autres techniques de programmation avances. Quatre exceptions : l'emploi de
 *          fonctions gnriques (template) pour la srialisation des enum et vector (a
 *          reste dans le connu pour leur utilisation puisque std::sort a t vue);
 *          l'emploi de l'appel direct du constructeur pour la cration des exceptions et
 *          dans EncoderXML; l'emploi de find_first_not_of pour la validation des noms de
 *          champs; l'emploi de unget() pour simplifier la lecture aprs les getline qui
 *          retirent un caractre de trop.
 *
 * Auteur : Michel Michaud (http://www.gdzid.com)
 * Cration : fvrier 2007
 */
#include <fstream>
#include <iomanip>
#include <string>
#include <cassert>
#include "UtilXML.h"
using namespace std;

namespace
{
// L'en-tte indique un jeu de caractres classiques et l'emploi d'un fichier seul (sans DTD)
const string EN_TTE_XML= "<?xml version=\"1.0\" encoding=\"ISO-8859-1\" standalone=\"yes\"?>";

// Le formatage est contrl par les quatre prochaines variables. Les deux premires sont
// modifiables grce aux fonctions DsactiverFormatage et ChangerIndentation. Parce qu'il
// n'y a qu'une seule occurrence des variables pour tout le code, le formatage ne
// fonctionnera normalement pas correctement si le programme fait la cration de plusieurs
// fichiers XML simultanment (DsactiverFormatage devrait alors tre appele).
bool g_formater= true; // Indique s'il faut indenter et faire des changements de ligne
int g_indentation= 3;  // Nombre d'espaces d'indentation  utiliser par niveau d'indentation
int g_niveauIndentation= 0; // Le niveau  utiliser pour la prochaine criture  formater
bool g_ouvert= false;  // Indique qu'une balise de dbut a t la dernire crite

/*****
 * NomChampValide : vrifie que le nom de champ est valide autant pour C++ (ce qui assure
 *     un peu qu'il n'y a pas de faute dans le nom du champ ou de la variable pass) que
 *     pour XML (ce qui est important en ralit). En plus de l'emploi des lettres,
 *     chiffres et soulign, on vrifie que le nom ne commence pas par "XML" ni variantes.
 */
 bool NomChampValide(const string& p_nomChamp) 
	{
	return !p_nomChamp.empty()                                     // Pas vide
		   && ! ('0' <= p_nomChamp[0] && p_nomChamp[0] <= '9')     // Pas de chiffre au dbut
		   && (string::npos == p_nomChamp.find_first_not_of(       // Tout est en caractres
				"_0123456789"                                      //  valides en gnral
		        "abcdefghijklmnopqrstuvwxyz"
		        "ABCDEFGHIJKLMNOPQRSTUVWXYZܟ"))
		   && ((p_nomChamp[0] != 'X' && p_nomChamp[0] != 'x')      // Ne commence pas par XML
		       || (p_nomChamp[1] != 'M' && p_nomChamp[1] != 'm')   //  (n'importe quelle
		       || (p_nomChamp[2] != 'L' && p_nomChamp[2] != 'l')); //  casse)
	}

string BaliseDbut(const string& p_nomChamp, const string& p_nomType)
	{
	assert(NomChampValide(p_nomChamp));
	return '<' + p_nomChamp + " type=\"" + p_nomType + "\">";
	}

string BaliseFin(const string& p_nomChamp)
	{
	assert(NomChampValide(p_nomChamp));
	return "</" + p_nomChamp + '>';
	}

string LireBalise(ifstream& p_es_fic)
	{
	p_es_fic >> ws;  // On saute les espaces et sauts de ligne (probablement notre formatage)
	string balise;                  // On veut la balise complte mais le getline ne le met
	getline(p_es_fic, balise, '>'); //  pas dans la variable. On l'ajoute donc manuellement
	return balise+'>';              //  pour le retour de la fonction.
	}

} // Fin du namespace anonyme

namespace UtilXML
{
void crireEntteXML(ofstream& p_es_fic)
	{
	p_es_fic << EN_TTE_XML;

	if (g_formater)
		p_es_fic << '\n';
	}

void VrifierEntteXML(ifstream& p_es_fic)
	{
	if (LireBalise(p_es_fic) != EN_TTE_XML)
		throw TypeErreurXML();
	}

void DsactiverFormatage()
	{
	g_formater= false;
	}

void ChangerIndentation(int p_indentation)
	{
	assert(p_indentation >= 0);
	g_indentation= p_indentation;
	g_formater= true;
	}

/*****
 * EncoderXML : cette version encode un seul caractre mais renvoie une string puisque
 *     l'encodage peut donner une rfrence d'entit comme "&lt;". La fonction remplace
 *     les caractres problmatiques (< et &) mais aussi trois autres (>, " et ') pour
 *     aider  la lisibilit du fichier XML.
 */
std::string EncoderXML(char p_caractre)
	{
	switch (p_caractre)
		{
		case '<' : return "&lt;"; break;
		case '&' : return "&amp;"; break;

		case '>'  : return "&gt;"; break;
		case '"'  : return "&quot;"; break;
		case '\'' : return "&apos;"; break;

		default  : return string(1, p_caractre); // Le caractre lui-mme, en string
		}
	}

std::string EncoderXML(const string& p_texte)
	{
	string texteEncod; // Le rsultat de l'encodage

	for (string::size_type i= 0; i != p_texte.size(); ++i)
		texteEncod+= EncoderXML(p_texte[i]);

	return texteEncod;
	}

/*****
 * DcoderXML : renvoie un texte o les caractres encods ont t remplacs par les
 *     caractres quivalents. Seuls les encodages des fonctions EncoderXML sont dcods.
 */
string DcoderXML(const string& p_texte)
	{
	string texteDcod; // Le rsultat du dcodage

	for (string::size_type i= 0; i != p_texte.size(); )
		{
		if (p_texte[i] != '&')
			{
			texteDcod+= p_texte[i];
			++i;
			}
		else
			{
			++i;
			// On va regarder ce qu'il y a aprs le &. Si on trouve un code connu, on va
			// mettre le caractre quivalent. Sinon on laisse le & tel quel.
			if (p_texte.substr(i, 3) == "lt;")
				{
				texteDcod+= '<';
				i+= 3;
				}
			else
			if (p_texte.substr(i, 4) == "amp;")
				{
				texteDcod+= '&';
				i+= 4;
				}
			else
			if (p_texte.substr(i, 3) == "gt;")
				{
				texteDcod+= '>';
				i+= 3;
				}
			else
			if (p_texte.substr(i, 5) == "quot;")
				{
				texteDcod+= '"';
				i+= 5;
				}
			else
			if (p_texte.substr(i, 5) == "apos;")
				{
				texteDcod+= '\'';
				i+= 5;
				}
			else
				texteDcod+= '&'; // On ne sait pas quoi en faire, on le laisse
			}
		}

	return texteDcod;
	}

void crireBaliseDbutStruct(ofstream& p_es_fic, 
							 const string& p_nomChamp, const string& p_nomType)
	{
	XmlcrireBaliseDbutBase(p_es_fic, p_nomChamp, "struct " + p_nomType);
	}

void VrifierBaliseDbutStruct(ifstream& p_es_fic,
							   const string& p_nomChamp, const string& p_nomType)
	{
	XmlVrifierBaliseDbutBase(p_es_fic, p_nomChamp, "struct " + p_nomType);
	}

/*****
 * crireBaliseDbutFichier : crit une balise de dbut (ouverture), pour le fichier XML
 *                            complet, avec formatage si dsir.
 */
void crireBaliseDbutFichier(std::ofstream& p_es_fic, 
					          const std::string& p_identification, 
							  const std::string& p_version)
	{
	assert(NomChampValide(p_identification)); // L'identification doit suivre les mmes
	                                          //  rgles que les noms de champs

	// En principe, c'est la premire balise (aprs l'en-tte XML) donc aucun formatage
	// ne devrait tre appliqu...
	if (g_formater)
		{
		if (g_ouvert)
			p_es_fic << '\n';
		else
			g_ouvert= true;

		p_es_fic << setw(g_indentation*g_niveauIndentation) << "";
		}

	++g_niveauIndentation; // On compte mme si on ne formate pas, au cas o on devrait
	                       //  formater plus tard, en plein milieu

	p_es_fic << '<' << p_identification << " version=\"" << p_version << "\">";
	}

void VrifierBaliseDbutFichier(std::ifstream& p_es_fic,
						       const std::string& p_identification,
							   const std::string& p_version)
	{
	string baliseDbut= LireBalise(p_es_fic);

	if (baliseDbut != ('<' + p_identification + " version=\"" + p_version + "\">"))
		throw TypeErreurXML();
	}


/*****
 * crireBaliseFin : crit une balise de fin (fermeture), avec formatage si dsir.
 */
void crireBaliseFin(ofstream& p_es_fic, const string& p_nom)
	{
	--g_niveauIndentation; // On compte mme si on ne formate pas, au cas o on devrait
	                       //  formater plus tard, en plein milieu
	if (g_formater)
		{
		if (g_ouvert)              // On indente la balise seulement si elle est au
			g_ouvert= false;       //  dbut d'une ligne donc aprs l'criture
		else                       //  d'une autre balise de fin
			p_es_fic << setw(g_indentation*g_niveauIndentation) << "";
		}

	p_es_fic << BaliseFin(p_nom);
	
	if (g_formater) // Toute balise de fin est suivi d'un changement de ligne
		p_es_fic << '\n';
	}

void VrifierBaliseFin(ifstream& p_es_fic, const string& p_nom)
	{
	string baliseFin= LireBalise(p_es_fic);

	if (baliseFin != BaliseFin(p_nom))
		throw TypeErreurXML();
	}

void SrialiserInt(ofstream& p_es_fic, const string& p_nomChamp, int p_valeur)
	{
	XmlcrireBaliseDbutBase(p_es_fic, p_nomChamp, "int");
	p_es_fic << p_valeur;
	crireBaliseFin(p_es_fic, p_nomChamp);
	}

void DsrialiserInt(ifstream& p_es_fic, const string& p_nomChamp, int& p_s_valeur)
	{
	XmlVrifierBaliseDbutBase(p_es_fic, p_nomChamp, "int");
	p_es_fic >> p_s_valeur;

	if (p_es_fic.fail())
		throw TypeErreurXML();

	VrifierBaliseFin(p_es_fic, p_nomChamp);
	}

void SrialiserLong(ofstream& p_es_fic, const string& p_nomChamp, long p_valeur)
	{
	XmlcrireBaliseDbutBase(p_es_fic, p_nomChamp, "long");
	p_es_fic << p_valeur;
	crireBaliseFin(p_es_fic, p_nomChamp);
	}

void DsrialiserLong(ifstream& p_es_fic, const string& p_nomChamp, long& p_s_valeur)
	{
	XmlVrifierBaliseDbutBase(p_es_fic, p_nomChamp, "long");
	p_es_fic >> p_s_valeur;

	if (p_es_fic.fail())
		throw TypeErreurXML();

	VrifierBaliseFin(p_es_fic, p_nomChamp);
	}

void SrialiserDouble(ofstream& p_es_fic, const string& p_nomChamp, double p_valeur)
	{
	XmlcrireBaliseDbutBase(p_es_fic, p_nomChamp, "double");
	p_es_fic << scientific << setprecision(16) << p_valeur;
	crireBaliseFin(p_es_fic, p_nomChamp);
	}

void DsrialiserDouble(ifstream& p_es_fic, const string& p_nomChamp, double& p_s_valeur)
	{
	XmlVrifierBaliseDbutBase(p_es_fic, p_nomChamp, "double");
	p_es_fic >> p_s_valeur;

	if (p_es_fic.fail())
		throw TypeErreurXML();

	VrifierBaliseFin(p_es_fic, p_nomChamp);
	}

void SrialiserBool(ofstream& p_es_fic, const string& p_nomChamp, bool p_valeur)
	{
	XmlcrireBaliseDbutBase(p_es_fic, p_nomChamp, "bool");
	p_es_fic << p_valeur;
	crireBaliseFin(p_es_fic, p_nomChamp);
	}

void DsrialiserBool(ifstream& p_es_fic, const string& p_nomChamp, bool& p_s_valeur)
	{
	XmlVrifierBaliseDbutBase(p_es_fic, p_nomChamp, "bool");
	p_es_fic >> p_s_valeur;

	if (p_es_fic.fail())
		throw TypeErreurXML();

	VrifierBaliseFin(p_es_fic, p_nomChamp);
	}

void SrialiserChar(ofstream& p_es_fic, const string& p_nomChamp, char p_valeur)
	{
	XmlcrireBaliseDbutBase(p_es_fic, p_nomChamp, "char");
	p_es_fic << EncoderXML(p_valeur);
	crireBaliseFin(p_es_fic, p_nomChamp);
	}

/*****
 * DsrialiserChar : extrait un caractre, qui peut avoir t encod, donc fait
 *                    en plus le dcodage au besoin.
 */
void DsrialiserChar(ifstream& p_es_fic, const string& p_nomChamp, char& p_s_valeur)
	{
	XmlVrifierBaliseDbutBase(p_es_fic, p_nomChamp, "char");
	string texte;
	getline(p_es_fic, texte, '<');

	if (p_es_fic.fail())
		throw TypeErreurXML();

	p_es_fic.unget();
	p_s_valeur= DcoderXML(texte)[0];
	VrifierBaliseFin(p_es_fic, p_nomChamp);
	}

void SrialiserString(ofstream& p_es_fic, const string& p_nomChamp, const string& p_valeur)
	{
	XmlcrireBaliseDbutBase(p_es_fic, p_nomChamp, "std::string");
	p_es_fic << EncoderXML(p_valeur);
	crireBaliseFin(p_es_fic, p_nomChamp);
	}

/*****
 * DsrialiserString : extrait du texte, pouvant contenir des caractres encods, donc fait
 *                      en plus le dcodage au besoin.
 */
void DsrialiserString(ifstream& p_es_fic, const string& p_nomChamp, string& p_s_valeur)
	{
	XmlVrifierBaliseDbutBase(p_es_fic, p_nomChamp, "std::string");
	getline(p_es_fic, p_s_valeur, '<');

	if (p_es_fic.fail())
		throw TypeErreurXML();

	p_es_fic.unget();
	p_s_valeur= DcoderXML(p_s_valeur);
	VrifierBaliseFin(p_es_fic, p_nomChamp);
	}

void SrialiserEnum(ofstream& p_es_fic, const string& p_nomChamp, const string& p_nomType,
		            long p_valeur, const string& p_valeurEnTexte)
	{
	XmlcrireBaliseDbutBase(p_es_fic, p_nomChamp, "enum " + p_nomType);
	p_es_fic << p_valeur
			 << ((p_valeurEnTexte.empty()) ? "" : ('('+p_valeurEnTexte+')'));
	crireBaliseFin(p_es_fic, p_nomChamp);
	}

////////////////////////////// LES FONCTIONS  PRIVES  ////////////////////////////////

/*****
 * XmlcrireBaliseDbutBase : fonction d'aide ( prive ) qui crit une balise de dbut
 *                            (ouverture), avec formatage si dsir.
 */
void XmlcrireBaliseDbutBase(ofstream& p_es_fic, 
							  const string& p_nomChamp, const string& p_nomType)
	{
	if (g_formater)
		{
		if (g_ouvert)
			p_es_fic << '\n';  // On change de ligne entre deux balises de dbut
		else                   //  conscutives (sinon on est dj au dbut d'une ligne)
			g_ouvert= true;

		p_es_fic << setw(g_indentation*g_niveauIndentation) << ""; // Indentation
		}

	++g_niveauIndentation; // On compte mme si on ne formate pas, au cas o on devrait
	                       //  formater plus tard, en plein milieu
	p_es_fic << BaliseDbut(p_nomChamp, p_nomType);
	}

/*****
 * XmlVrifierBaliseDbutBase : fonction d'aide ( prive ) qui vrifie une balise de
 *                              dbut (ouverture).
 */
void XmlVrifierBaliseDbutBase(ifstream& p_es_fic, 
								const string& p_nomChamp, const string& p_nomType)
	{
	string baliseDbut= LireBalise(p_es_fic);

	if (baliseDbut != BaliseDbut(p_nomChamp, p_nomType))
		throw TypeErreurXML();
	}

/*****
 * XmlDsrialiserEnumEnLong : fonction d'aide ( prive ) pour la dsrialisation des
 *     types par enumration; elle extrait la valeur numrique sous la forme d'un long.
 */
void XmlDsrialiserEnumEnLong(ifstream& p_es_fic,
		   			           const string& p_nomChamp, const string& p_nomType,
		                       long& p_s_valeur)
	{
	XmlVrifierBaliseDbutBase(p_es_fic, p_nomChamp, "enum " + p_nomType);
	p_es_fic >> p_s_valeur;

	// On va maintenant lire la valeur texte dans une variable. On aurait pu simplement
	// sauter le texte avec ignore puisqu'on ne l'utilisera pas. Par contre, la variable
	// permet de simplifier le dboguage au besoin (on peut consulter la variable).
	string valeurTexte;
	getline(p_es_fic, valeurTexte, '<');

	if (p_es_fic.fail())
		throw TypeErreurXML();

	p_es_fic.unget();
	VrifierBaliseFin(p_es_fic, p_nomChamp);
	}

}
