/* UtilXML.h : dclaration des types et fonctions d'assistance  la srialisation en
 *     format XML. Voir les commentaires ou la documentation pour plus de dtails
 *     sur l'utilisation ou la structure des fichiers crs.
 *
 * Auteur : Michel Michaud (http://www.gdzid.com)
 * Cration : fvrier 2007
 */
#if !defined UTILXML_H_INCLUS
#define UTILXML_H_INCLUS
#include <string>
#include <fstream>
#include <vector>
#include <cassert>

namespace UtilXML
{
// Type de l'exception leve lors de la dtection d'une erreur.
struct TypeErreurXML {};

// Fonctions pour l'en-tte XML standard (<?xml version='1.0'...)
// On peut ou non les employer, mais leur emploi rend le fichier conforme ce qui permet
// de l'ouvrir facilement dans Internet Explorer ou Visual Studio par exemple.
void crireEntteXML(std::ofstream& p_es_fic);
void VrifierEntteXML(std::ifstream& p_es_fic);

// Fonctions permettant de changer le formatage appliqu par dfaut par les autres
// fonctions. On peut dsactiver compltement le formatage ou changer l'indentation
// (qui est  3 par dfaut). ChangerIndentation ractive aussi le formatage.
void DsactiverFormatage();
void ChangerIndentation(int p_indentation);

// Fonctions d'encodage/dcodage qui remplacent les caractres problmatiques < et &
// par les codes quivalents (&lt; et &amp;). Elles modifient aussi les trois caractres
// >"' par &gt; &quot; et &apos; pour aider  la lisibilit. En principe, il ne devrait
// pas tre ncessaire de les utiliser explicitement, car les fonctions SrialiserChar,
// DsrialiserChar, SrialiserString et DsrialiserString les utilisent dj.
std::string EncoderXML(char p_caractre);
std::string EncoderXML(const std::string& p_texte);
std::string DcoderXML(const std::string& p_texte);

// Fonctions pour l'criture et vrification des balises de dbut pour la cration
// des fonctions de srialisation/dsrialisation des types struct personnels.
// La balise est de la forme <nomChamp type="struct nomType">.
void crireBaliseDbutStruct(std::ofstream& p_es_fic, 
					         const std::string& p_nomChamp, const std::string& p_nomType);
void VrifierBaliseDbutStruct(std::ifstream& p_es_fic,
						       const std::string& p_nomChamp, const std::string& p_nomType);

// Fonctions pour l'criture et vrification des balises de dbut pour la cration
// des fonctions gnrales de srialisation (donc du fichier au complet).
// La balise est de la forme <identification version="version">.
void crireBaliseDbutFichier(std::ofstream& p_es_fic, 
					          const std::string& p_identification, 
							  const std::string& p_version);
void VrifierBaliseDbutFichier(std::ifstream& p_es_fic,
						       const std::string& p_identification,
							   const std::string& p_version);

// Fonctions pour l'criture et vrification des balises de fin pour la cration
// des fonctions de srialisation des struct et de srialisation gnrale.
// La balise est de la forme </nom>.
void crireBaliseFin(std::ofstream& p_es_fic, const std::string& p_nom);
void VrifierBaliseFin(std::ifstream& p_es_fic, const std::string& p_nom);

// Fonctions pour la srialisation/dsrialisation des types de base (et std::string).
// La forme srialise est <nomChamp type="nomType">valeur</nomChamp>
void SrialiserInt(std::ofstream& p_es_fic, const std::string& p_nomChamp, 
				   int p_valeur);
void DsrialiserInt(std::ifstream& p_es_fic, const std::string& p_nomChamp,
					 int& p_s_valeur);
void SrialiserLong(std::ofstream& p_es_fic, const std::string& p_nomChamp, 
					long p_valeur);
void DsrialiserLong(std::ifstream& p_es_fic, const std::string& p_nomChamp,
					  long& p_s_valeur);
void SrialiserDouble(std::ofstream& p_es_fic, const std::string& p_nomChamp,
					  double p_valeur);
void DsrialiserDouble(std::ifstream& p_es_fic, const std::string& p_nomChamp,
						double& p_s_valeur);
void SrialiserBool(std::ofstream& p_es_fic, const std::string& p_nomChamp,
					bool p_valeur);
void DsrialiserBool(std::ifstream& p_es_fic, const std::string& p_nomChamp,
					  bool& p_s_valeur);
void SrialiserChar(std::ofstream& p_es_fic, const std::string& p_nomChamp,
					char p_valeur);
void DsrialiserChar(std::ifstream& p_es_fic, const std::string& p_nomChamp,
					  char& p_s_valeur);
void SrialiserString(std::ofstream& p_es_fic, const std::string& p_nomChamp, 
					  const std::string& p_valeur);
void DsrialiserString(std::ifstream& p_es_fic, const std::string& p_nomChamp,
						std::string& p_s_valeur);

//******** Trois fonctions  prives ,  ne pas utiliser directement.
void XmlcrireBaliseDbutBase(std::ofstream&, const std::string&, const std::string&);
void XmlVrifierBaliseDbutBase(std::ifstream&, const std::string&, const std::string&);
void XmlDsrialiserEnumEnLong(std::ifstream&, const std::string&, const std::string&, long&);
//********

// Fonction pour la srialisation des types enum (le paramtre est de type long, mais la
// conversion est automatique). On peut passer ou non la valeur en texte en dernier
// paramtre : une chane vide tel le paramtre par dfaut fera crire la valeur numrique
// seulement, sinon la valeur en texte est crite entre parenthses aprs la valeur
// numrique. Le mot  enum  est automatiquement ajout au nom du type. La forme
// srialise est <nomChamp type="enum nomType">valeurNumrique(valeurEnTexte)</nomChamp>
// ou <nomChamp type="enum nomType">valeurNumrique</nomChamp>
void SrialiserEnum(std::ofstream& p_es_fic, 
					const std::string& p_nomChamp, const std::string& p_nomType,
		            long p_valeur, const std::string& p_valeurEnTexte= "");

// Fonction pour la dsrialisation des types enum (il s'agit d'une fonction gnrique pour
// permettre le paramtre de sortie d'un type enum quelconque).
template <typename T>
inline
void DsrialiserEnum(std::ifstream& p_es_fic,
				      const std::string& p_nomChamp, const std::string& p_nomType, 
		              T& p_s_valeur)
	{
	long valeur; // La valeur numrique de la valeur du type enum srialise
	XmlDsrialiserEnumEnLong(p_es_fic, p_nomChamp, p_nomType, valeur);
	p_s_valeur= static_cast<T>(valeur);
	}

// Fonctions pour la srialisation/dsrialisation des std::vector. Il s'agit de fonctions
// gnriques pour permettre le passage en paramtre d'un vecteur de donnes de n'importe
// quel type ainsi que le passage d'une fonction approprie pour la srialisation ou
// dsrialisation des lments. La forme srialise est
// <nomChamp type="std::vector"
//    <taille type="long">taille du vecteur</taille>
//    <lments type="nomType">
//       <lment type="nomType">valeur du vecteur[0]</lment>
//       <lment type="nomType">valeur du vecteur[1]</lment>
//       ...
//    </lments>
// </nomChamp>
template <typename T, typename F>
inline
void SrialiserVector(std::ofstream& p_es_fic,
					  const std::string& p_nomChamp, const std::string& p_nomTypeDeslments, 
		              const std::vector<T>& p_vecteur,
					  F p_Srialiser)
	{
	XmlcrireBaliseDbutBase(p_es_fic, p_nomChamp, "std::vector");
	SrialiserLong(p_es_fic, "taille", p_vecteur.size());
	XmlcrireBaliseDbutBase(p_es_fic, "lments", p_nomTypeDeslments);

	for (int i= 0; i != p_vecteur.size(); ++i)
		p_Srialiser(p_es_fic, "lment", p_vecteur[i]);

	crireBaliseFin(p_es_fic, "lments");
	crireBaliseFin(p_es_fic, p_nomChamp);
	}

template <typename T, typename F>
inline
void DsrialiserVector(std::ifstream& p_es_fic,
					    const std::string& p_nomChamp, const std::string& p_nomTypeDeslments,
		                std::vector<T>& p_s_vecteur,
					    F p_Dsrialiser)
	{
	XmlVrifierBaliseDbutBase(p_es_fic, p_nomChamp, "std::vector");
	
	long taille; // La taille du vecteur  dsrialiser
	DsrialiserLong(p_es_fic, "taille", taille);
	p_s_vecteur.resize(taille);

	XmlVrifierBaliseDbutBase(p_es_fic, "lments", p_nomTypeDeslments);

	for (int i= 0; i != p_s_vecteur.size(); ++i)
		p_Dsrialiser(p_es_fic, "lment", p_s_vecteur[i]);

	VrifierBaliseFin(p_es_fic, "lments");
	VrifierBaliseFin(p_es_fic, p_nomChamp);
	}
}
#endif
